// Doze.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include <string>

#define SUSPEND_SLEEP TRUE
#define SUSPEND_HIBERNATE FALSE

const LONGLONG FILETIME_UNITS_PER_MICROSECOND = 10;
const LONGLONG FILETIME_UNITS_PER_MILLISECOND = FILETIME_UNITS_PER_MICROSECOND * 1000;
const LONGLONG FILETIME_UNITS_PER_SECOND = FILETIME_UNITS_PER_MILLISECOND * 1000;

//using System.Runtime.InteropServices; //to DllImport

const int MONITOR_ON = -1;
const int MONITOR_OFF = 2;
const int MONITOR_STANBY = 1;

struct NAPTIME_PARAMS
{
	BOOL fSuspend;
	BOOL fForce;
};

DWORD WINAPI NapTime(LPVOID lpParameter)
{
	NAPTIME_PARAMS *pParams = (NAPTIME_PARAMS *)lpParameter;

	// Sleep for just long enough for the other thread to have started waiting.
	_tprintf(_T("Setting system power state, dozing now.\n"));
	::Sleep(20);
	::SetSystemPowerState(pParams->fSuspend, pParams->fForce);

	return 0;
}

int Error(const char *message, DWORD error)
{
	LPSTR pBuffer;
	DWORD buffer_len = ::FormatMessageA(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ALLOCATE_BUFFER, NULL, error, 0, (LPSTR)&pBuffer, 0, NULL);
	if (buffer_len != 0)
	{
		// Remove trailing CRLF.
		pBuffer[buffer_len - 2] = 0;

		fprintf(stderr, "%s failed: %s\n", message, pBuffer);
		::LocalFree(pBuffer);

		return error;
	}
	else
	{
		fprintf(stderr, "%s failed: %d\n", message, error);
		return error;
	}
}

void ShowUsage()
{
		_tprintf(_T("DOZE - Put the system to sleep or hibernate, once, then wake up.\n\n"));
		_tprintf(_T("Usage:\n"));
		_tprintf(_T("Doze-vc9.exe [3|4] [secs]\n\n"));
		_tprintf(_T("Parameter 1 (required): Power state.\n"));
		_tprintf(_T("                        Use 3 for a sleep/suspend and 4 for a hibernate.\n\n"));
		_tprintf(_T("Parameter 2 (optional): Seconds to sleep or hibernate.\n"));
		_tprintf(_T("                        If not specified, it will default to 30 seconds\n"));
		_tprintf(_T("                        for sleep and 60 seconds for hibernate.\n"));
		_tprintf(_T("                        Note: the minimum sleep value is 30 seconds, and\n"));
		_tprintf(_T("                        the minimum hibernate value is 60 seconds.\n\n"));
		_tprintf(_T("Example:\n\n"));
		_tprintf(_T("Doze-vc9.exe 3 45\n"));
		_tprintf(_T("Computer will go into sleep/suspend mode for 45 seconds then wake up again.\n"));

		return;
}

int _tmain(int argc, _TCHAR* argv[])
{
	BOOL fForce = FALSE;
	BOOL fSuspend = SUSPEND_SLEEP;
	DWORD delaySeconds = 30;

	if (argc <2 || argc >3)
	{
		_tprintf(_T("Wrong number of arguments.\n\n"));
		ShowUsage();
		return 0;
	}

	int nTempSuspendType;

	nTempSuspendType = _ttoi(argv[1]);



	switch (nTempSuspendType)
	{
	  case 3:
		_tprintf(_T("Doze: Preparing to SLEEP the computer.\n"));
		fSuspend = SUSPEND_SLEEP;
		delaySeconds = 30;
		break;
	  case 4:
		_tprintf(_T("Doze: Preparing to HIBERNATE the computer.\n"));
		fSuspend = SUSPEND_HIBERNATE;
		delaySeconds = 60;
		break;
	  default:
		_tprintf(_T("Incorrect arguments specified.\n\n"));
		ShowUsage();
		return 0;
		break;
	}

	if (argc == 3)
	{
		delaySeconds = _ttoi(argv[2]);
		if (delaySeconds < 30) delaySeconds = 30;
		if (delaySeconds < 60 && fSuspend == SUSPEND_HIBERNATE) delaySeconds = 60;
		_tprintf(_T("Doze: System will doze for %d seconds.\n"), delaySeconds);
	}

	HANDLE hToken;
	if (!::OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES, &hToken))
		return Error("OpenProcessToken failed", ::GetLastError());

	LUID privilegeId;
	if (!::LookupPrivilegeValue(NULL, SE_SHUTDOWN_NAME, &privilegeId))
		return Error("LookupPrivilegeValue failed", ::GetLastError());

	TOKEN_PRIVILEGES privs;
	privs.PrivilegeCount = 1;
	privs.Privileges[0].Luid = privilegeId;
	privs.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;

	if (!::AdjustTokenPrivileges(hToken, FALSE, &privs, sizeof(TOKEN_PRIVILEGES), NULL, 0))
		return Error("AdjustTokenPrivileges failed", ::GetLastError());

	HANDLE hTimer = ::CreateWaitableTimer(NULL, FALSE, NULL);
	if (hTimer == NULL)
		return Error("CreateWaitableTimer failed", ::GetLastError());

	SYSTEMTIME systemTime;
	::GetSystemTime(&systemTime);

	TCHAR timeStr[200];
	::GetTimeFormat(LOCALE_SYSTEM_DEFAULT, 0, &systemTime, NULL, timeStr, ARRAYSIZE(timeStr));
	_tprintf(_T("It is now %s UTC\n"), timeStr);

	FILETIME fileTime;
	::SystemTimeToFileTime(&systemTime, &fileTime);

	LARGE_INTEGER wakeTime;
	wakeTime.LowPart = fileTime.dwLowDateTime;
	wakeTime.HighPart = fileTime.dwHighDateTime;

	wakeTime.QuadPart += delaySeconds * FILETIME_UNITS_PER_SECOND;

	_tprintf(_T("Dozing in 3... ")); Sleep (1000);
	_tprintf(_T("2... ")); Sleep (1000);
	_tprintf(_T("1... \n")); Sleep (1000);

	_tprintf(_T("Setting doze alarm for %d seconds\n"), delaySeconds);
	if (::SetWaitableTimer(hTimer, &wakeTime, 0, NULL, NULL, TRUE))
	{
		if (GetLastError() == ERROR_NOT_SUPPORTED)
			return Error("Using waitable timers to resume from sleep is not supported on this platform.", ERROR_NOT_SUPPORTED);
	}
	else
	{
		return Error("SetWaitableTimer failed", ::GetLastError());
	}
	
	// We need to wait on the timer, otherwise Windows won't bother resuming.
	// But we can't go to sleep *and* wait at the same time, so we'll need a background thread.
	// But that introduces a race condition by itself -- if it calls SetSystemPowerState before
	// we start waiting, we're back where we started, so we introduce a small sleep to get everything
	// in the right order.
	
	NAPTIME_PARAMS params = { fSuspend, fForce };
	::QueueUserWorkItem(NapTime, &params, WT_EXECUTEDEFAULT);
	::WaitForSingleObject(hTimer, INFINITE);

	// Fix issue where the monitor is turned off when it comes back from suspend.
	SetThreadExecutionState(ES_DISPLAY_REQUIRED);
	
	_tprintf(_T("Resuming from doze.\n"));

	// To prevent it from going back to sleep...
	::keybd_event(VK_NONAME, 0, 0, 0);


	return 0;
}

